@@ -19,8 +19,6 @@
 #define IF_ALLOCATE_ERR \
 	if (PAGESIZE*(n - m) > pagetochar(MAXPAGE) - core_end)
 
-/* smallest address for a page table entry */
-#define DBEGIN 0x80800000
 /* Begin for cmpinclude */
 
 
