@@ -360,8 +360,8 @@ struct ustring {
 
 };
 
-#define USHORT(x,i) (((unsigned short *)(x)->ust.ust_self)[i])
-#define SHORT(x,i) ((( short *)(x)->ust.ust_self)[i])
+#define USHORT_GCL(x,i) (((unsigned short *)(x)->ust.ust_self)[i])
+#define SHORT_GCL(x,i) ((( short *)(x)->ust.ust_self)[i])
 
 #define BV_OFFSET(x) ((type_of(x)==t_bitvector ? x->bv.bv_offset : \
 		       type_of(x)== t_array ? x->a.a_offset : (abort(),0)))
@@ -455,7 +455,7 @@ struct s_data {object name;
 
 #define S_DATA(x) ((struct s_data *)((x)->str.str_self))
 #define SLOT_TYPE(def,i) (((S_DATA(def))->raw->ust.ust_self[i]))
-#define SLOT_POS(def,i) USHORT(S_DATA(def)->slot_position,i)
+#define SLOT_POS(def,i) USHORT_GCL(S_DATA(def)->slot_position,i)
 #define STREF(type,x,i) (*((type *)(((char *)((x)->str.str_self))+(i))))
 
 
@@ -742,14 +742,14 @@ struct typemanager {
 	short   tm_nppage;	/*  number per page  */
 	object	tm_free;	/*  free list  */
 				/*  Note that it is of type object.  */
-	int	tm_nfree;	/*  number of free elements  */
-	int	tm_nused;	/*  number of elements used  */
-	int	tm_npage;	/*  number of pages  */
-	int	tm_maxpage;	/*  maximum number of pages  */
+	long	tm_nfree;	/*  number of free elements  */
+	long	tm_nused;	/*  number of elements used  */
+	long	tm_npage;	/*  number of pages  */
+	long	tm_maxpage;	/*  maximum number of pages  */
 	char	*tm_name;	/*  type name  */
 	int	tm_gbccount;	/*  GBC count  */
 	object  tm_alt_free;    /*  Alternate free list (swap with tm_free) */
-	int     tm_alt_nfree;   /*  Alternate nfree (length of nfree) */
+	long     tm_alt_nfree;   /*  Alternate nfree (length of nfree) */
 	short   tm_sgc;         /*  this type has at least this many
 				    sgc pages */
 	short   tm_sgc_minfree;   /* number free on a page to qualify for
@@ -759,7 +759,8 @@ struct typemanager {
 	short   tm_max_grow;    /* max amount to grow when growing */
 	short   tm_growth_percent;  /* percent to increase maxpages */
 	short   tm_percent_free;  /* percent which must be free after a gc for this type */
-
+        short   tm_distinct;       /* pages of this type are distinct */
+        float   tm_adjgbccnt;
 };
 
 
@@ -784,19 +785,33 @@ struct contblock {		/*  contiguous block header  */
 */
 EXTER struct contblock *cb_pointer;	/*  contblock pointer  */
 
+/* SGC cont pages: After SGC_start, old_cb_pointer will be a linked
+   list of free blocks on non-SGC pages, and cb_pointer will be
+   likewise for SGC pages.  CM 20030827*/
+EXTER struct contblock *old_cb_pointer;	/*  old contblock pointer when in SGC  */
+
+/* SGC cont pages: FIXME -- at some point, enable runtime disabling of
+   SGC cont pages.  Right now, the tm_sgc variable for type contiguous
+   will govern only the possible attempt to get new pages for SGC.
+   Contiguous pages normally allocated when SGC is on will always be
+   marked with SGC_PAGE_FLAG, as the current GBC algorithm always uses
+   sgc_contblock_sweep_phase in this case. */
+/* #define SGC_CONT_ENABLED (sgc_enabled && tm_table[t_contiguous].tm_sgc) */
+#define SGC_CONT_ENABLED (sgc_enabled)
+
 /*
 	Variables for memory management.
 */
-EXTER int ncb;			/*  number of contblocks  */
+EXTER long ncb;			/*  number of contblocks  */
 /* int ncbpage;			  number of contblock pages  */
 #define ncbpage tm_table[t_contiguous].tm_npage
-#define maxcbpage   tm_table[t_contiguous].tm_maxpage
-#define cbgbccount tm_table[t_relocatable].tm_gbccount  
+#define maxcbpage tm_table[t_contiguous].tm_maxpage
+#define cbgbccount tm_table[t_contiguous].tm_gbccount  
   
 
 /* int maxcbpage; maximum number of contblock pages  */
 EXTER 
-int holepage;			/*  hole pages  */
+long holepage;			/*  hole pages  */
 #define nrbpage tm_table[t_relocatable].tm_npage
 #define rbgbccount tm_table[t_relocatable].tm_gbccount
 /* int nrbpage;			  number of relblock pages  */
@@ -829,14 +844,19 @@ char *tmp_alloc;
   (ALLOC_ALIGNED(f,(n)*sizeof(type),sizeof(type)))
 
 
+/* FIXME  Make all other page constants scale similarly by default. */
 #ifndef HOLEPAGE
-#define	HOLEPAGE	128
+#define	HOLEPAGE	(MAXPAGE/10)
 #endif
 
 
-#define	INIT_HOLEPAGE	150
-#define	INIT_NRBPAGE	50
-#define	RB_GETA		512
+/* #define	INIT_HOLEPAGE	150 */
+/* #define	INIT_NRBPAGE	50 */
+/* #define	RB_GETA		512 */
+
+#define	INIT_HOLEPAGE	(6*HOLEPAGE/5)
+#define	INIT_NRBPAGE	(INIT_HOLEPAGE/3)
+#define	RB_GETA		(10*INIT_NRBPAGE)
 
 
 #ifdef AV
