@@ -10,8 +10,8 @@
   (MAYBE_DATA_P(pp) &&  ((char *)(pp) < heap_end))
 
 #ifndef page
-#define page(p)	(((long)(((char *)(p))-DBEGIN)>>PAGEWIDTH))
-#define	pagetochar(x)	((char *)(((x) << PAGEWIDTH) + DBEGIN))
+#define page(p)	((unsigned long)(((unsigned long)(((char *)(p))-DBEGIN)>>PAGEWIDTH)))
+#define	pagetochar(x)	((char *)((((unsigned long)x) << PAGEWIDTH) + DBEGIN))
 #endif
   
 #ifdef UNIX
@@ -23,12 +23,23 @@
 
 /* alignment required for pointers */
 #ifndef PTR_ALIGN
-#define PTR_ALIGN (sizeof(long *))
+#define PTR_ALIGN SIZEOF_LONG
 #endif
 
 #define ROUND_UP_PTR(n)	(((long)(n) + (PTR_ALIGN-1)) & ~(PTR_ALIGN-1))
 #define ROUND_DOWN_PTR(n) (((long)(n)  & ~(PTR_ALIGN-1)))
 
+/* alignment required for contiguous pointers */
+#if PTR_ALIGN < SIZEOF_CONTBLOCK
+#define CPTR_ALIGN SIZEOF_CONTBLOCK
+#else
+#define CPTR_ALIGN PTR_ALIGN
+#endif
+/* #define CPTR_ALIGN (PTR_ALIGN < sizeof(struct contblock) ? sizeof(struct contblock) : PTR_ALIGN) */
+
+#define ROUND_UP_PTR_CONT(n)	(((long)(n) + (CPTR_ALIGN-1)) & ~(CPTR_ALIGN-1))
+#define ROUND_DOWN_PTR_CONT(n) (((long)(n)  & ~(CPTR_ALIGN-1)))
+
 
 #ifdef SGC
 
@@ -48,14 +59,14 @@ char sgc_type_map[MAXPAGE];
 #define SGC_WRITABLE  (SGC_PERM_WRITABLE | SGC_TEMP_WRITABLE)
 #define SGC_PAGE (SGC_TEMP_WRITABLE | SGC_PAGE_FLAG)
 
-#define SGC_PAGE_P(p) (sgc_type_map[p] & SGC_PAGE_FLAG)
-#define WRITABLE_PAGE_P(p) (sgc_type_map[p] & SGC_WRITABLE)
-#define ON_SGC_PAGE(x) (sgc_type_map[page(x)] & SGC_PAGE_FLAG)
-#define ON_WRITABLE_PAGE(x) (sgc_type_map[page(x)] & SGC_WRITABLE)
+#define SGC_PAGE_P(p) ((unsigned long)p<MAXPAGE && (sgc_type_map[p] & SGC_PAGE_FLAG))
+#define WRITABLE_PAGE_P(p) ((unsigned long)p<MAXPAGE && (sgc_type_map[p] & SGC_WRITABLE))
+#define ON_SGC_PAGE(x) ((unsigned long)page(x)<MAXPAGE && (sgc_type_map[page(x)] & SGC_PAGE_FLAG))
+#define ON_WRITABLE_PAGE(x) ((unsigned long)page(x)<MAXPAGE && (sgc_type_map[page(x)] & SGC_WRITABLE))
 
 
-#define  IF_WRITABLE(x,if_code) do {int xSG= page(x); \
-			    if(((xSG & (-MAXPAGE)) ==0) && \
+#define  IF_WRITABLE(x,if_code) do {unsigned long xSG= page(x); \
+			    if(((xSG & (-MAXPAGE)) ==0) && xSG < MAXPAGE && \
 			       (sgc_type_map[xSG] & SGC_WRITABLE)) \
 				 {if_code;}} while(0)
 
@@ -78,8 +89,8 @@ enum sgc_type { SGC_NORMAL,   /* not allocated since the last sgc */
 
 #define TM_BASE_TYPE_P(i) (((int) (tm_table[i].tm_type)) == i)
 
-void perm_writable() ;
-void make_writable();
+/* void perm_writable(char *,long) ; */
+/* void make_writable(long,long); */
 
 
 #define ROUND_DOWN_PAGE_NO(x) ((x) - (x % page_multiple))
@@ -130,6 +141,3 @@ extern int page_multiple;
 */
 char type_map[MAXPAGE];
 
-
-
-
