@@ -93,10 +93,10 @@ int reserve_pages_for_signal_handler =30;
    reserve_pages_for_signal_handler pages on hand in the hole
  */
 char *
-alloc_page(int n)
+alloc_page(long n)
 {
 	char *e;
-	int m;
+	long m;
 	e = heap_end;
 	if (n >= 0) {
 		if (n >=
@@ -106,8 +106,11 @@ alloc_page(int n)
 		    {
 			holepage = new_holepage + n;
 
-			{int in_sgc=sgc_enabled;
+			{
+#ifdef SGC
+			  int in_sgc=sgc_enabled;
 			 if (in_sgc) sgc_quit();
+#endif
 			if(in_signal_handler)
 			  {fprintf(stderr,
 				   "Cant do relocatable gc in signal handler. \
@@ -115,12 +118,14 @@ Try to allocate more space to save for allocation during signals: \
 eg to add 20 more do (si::set-hole-size %ld %d)\n...start over ", new_holepage, 20+ reserve_pages_for_signal_handler); fflush(stderr); exit(1);}
 
 			GBC(t_relocatable);
+#ifdef SGC
 			if (in_sgc)
 			  {sgc_start();
 			   /* starting sgc can use up some pages
 			      and may move heap end, so start over
 			    */
 			   return alloc_page(n);}
+#endif
 		       }
 		}
 		holepage -= n;
@@ -141,11 +146,11 @@ eg to add 20 more do (si::set-hole-size %ld %d)\n...start over ", new_holepage,
 	if (n <= m)
 		return(e);
 
+
 	IF_ALLOCATE_ERR error("Can't allocate.  Good-bye!");
 #ifdef SGC
 	if (sgc_enabled)
 	  make_writable(page(core_end),page(core_end)+n-m);
-
 #endif	
 	core_end += PAGESIZE*(n - m);
 	return(e);}
@@ -154,19 +159,22 @@ eg to add 20 more do (si::set-hole-size %ld %d)\n...start over ", new_holepage,
 void
 add_page_to_freelist(char *p, struct typemanager *tm)
 {short t,size;
- int i=tm->tm_nppage,fw;
+ long i=tm->tm_nppage,fw;
+ unsigned long np=page(p);
  object x,f;
+ if (np>=MAXPAGE)
+   error("Address passed to add_page_to_freelist beyone MAXPAGES\n");
  t=tm->tm_type;
 #ifdef SGC
  {
-   int nn=page(p);
+   long nn=np;
    if (sgc_enabled) {
      if (!WRITABLE_PAGE_P(nn)) 
        make_writable(nn,nn+1);
    }
  }
 #endif
- type_map[page(p)]= t;
+ type_map[np]= t;
  size=tm->tm_size;
  f=tm->tm_free;
  x= (object)p;
@@ -175,7 +183,7 @@ add_page_to_freelist(char *p, struct typemanager *tm)
 #ifdef SGC
  if (sgc_enabled && tm->tm_sgc)
    {x->d.s=SGC_RECENT;
-    sgc_type_map[page(x)] = (SGC_PAGE_FLAG | SGC_TEMP_WRITABLE);}
+    sgc_type_map[np] = (SGC_PAGE_FLAG | SGC_TEMP_WRITABLE);}
  else x->d.s = SGC_NORMAL;
  
  /* array headers must be always writable, since a write to the
@@ -184,7 +192,7 @@ add_page_to_freelist(char *p, struct typemanager *tm)
     but just SGC_TOUCH the header each time you write to it.   this
     is what is done with t_structure */
   if (t== (tm_of(t_array)->tm_type))
-   sgc_type_map[page(x)] |= SGC_PERM_WRITABLE;
+   sgc_type_map[np] |= SGC_PERM_WRITABLE;
    
 #endif 
  fw= *(int *)x;
@@ -212,8 +220,110 @@ call_after_gbc_hook(t)
     }
 }
 
-#define PERCENT_FREE(tm)  ((tm->tm_percent_free ? tm->tm_percent_free : 10)/100.0)
+/* #define PERCENT_FREE(tm)  ((tm->tm_percent_free ? tm->tm_percent_free : 10)/100.0) */
+#define PERCENT_FREE(tm)  ((tm->tm_percent_free ? tm->tm_percent_free : 30)/100.0)
 
+static int
+grow_linear(int old, int fract, int grow_min, int grow_max) {
+  
+  int delt;
+  if (fract==0) 
+    fract=50;
+  if (grow_min==0) 
+    grow_min=1;
+  if (grow_max==0) 
+/*     grow_max=1000; */
+    grow_max=0.1*MAXPAGE;
+
+  delt=(old*fract)/100;
+  delt= (delt < grow_min ? grow_min:
+	 delt > grow_max ? grow_max:
+	 delt);
+
+  return old + delt;
+
+}
+
+/* GCL's traditional garbage collecting algorithm placed heavy emphasis
+   on conserving memory.  Maximum page allocations of each object type
+   were only increased when the objects in use after GBC exceeded a
+   certain percentage threshold of the current maximum.  This allowed
+   a situation in which a growing heap would experience significant
+   performance degradation due to GBC runs triggered by types making
+   only temporary allocations -- the rate of GBC calls would be
+   constant while the cost for each GBC would grow with the size of
+   the heap.
+
+   We implement here a strategy designed to approximately optimize the
+   product of the total GBC call rate times the cost or time taken for
+   each GBC.  The rate is approximated from the actual gbccounts so
+   far experienced, while the cost is taken to be simply proportional
+   to the heap size at present.  This can be further tuned by taking
+   into account the number of pointers in each object type in the
+   future, but at present objects of several different types but
+   having the same size are grouped together in the type manager
+   table, so this step becomes more involved.
+
+   After each GBC, we calculate the maximum of the function
+   (gbc_rate_other_types + gbc_rate_this_type *
+   current_maxpage/new_maxpage)*(sum_all_maxpages-current_maxpage+new_maxpage).
+   If the benefit in the product from adopting the new_maxpage is
+   greater than 5%, we adopt it, and adjust the gbccount for the new
+   basis.  Corrections are put in place for small GBC counts, and the
+   possibility that GBC calls of only a single type are ever
+   triggered, in which case the optimum new_maxpage would diverge in
+   the simple analysis above.
+
+   20040403 CM */
+
+DEFVAR("*OPTIMIZE-MAXIMUM-PAGES*",sSAoptimize_maximum_pagesA,SI,sLt,"");
+#define OPTIMIZE_MAX_PAGES (sSAoptimize_maximum_pagesA ==0 || sSAoptimize_maximum_pagesA->s.s_dbind !=sLnil) 
+DEFVAR("*NOTIFY-OPTIMIZE-MAXIMUM-PAGES*",sSAnotify_optimize_maximum_pagesA,SI,sLnil,"");
+#define MMAX_PG(a_) ((a_)->tm_type == t_relocatable ? (a_)->tm_npage : (a_)->tm_maxpage)
+static long
+opt_maxpage(struct typemanager *my_tm) {
+
+  double x=0.0,y=0.0,z,r;
+  long mmax_page;
+  struct typemanager *tm,*tme;
+  long mro=0,tro=0;
+
+  for (tm=tm_table,tme=tm+sizeof(tm_table)/sizeof(*tm_table);tm<tme;tm++) {
+    x+=tm->tm_adjgbccnt;
+    y+=MMAX_PG(tm);
+  }
+  mmax_page=MMAX_PG(my_tm);
+#if 0
+  if (sgc_enabled) {
+    y-=(tro=sgc_count_read_only_type(-1));
+    mmax_page-=(mro=sgc_count_read_only_type(my_tm->tm_type));
+  }
+#endif
+
+  z=my_tm->tm_adjgbccnt-1;
+  z/=(1+x-0.9*my_tm->tm_adjgbccnt);
+  z*=(y-mmax_page)*mmax_page;
+  z=sqrt(z);
+
+  if (z<=mmax_page)
+    return 0;
+
+  r=((x-my_tm->tm_adjgbccnt)+ my_tm->tm_adjgbccnt*mmax_page/z)*(y-mmax_page+z);
+  r/=x*y;
+  if (sSAnotify_optimize_maximum_pagesA->s.s_dbind!=sLnil)
+    printf("[type %u max %lu(%lu) opt %u   y %u(%lu) gbcrat %f sav %f]\n",
+	   my_tm->tm_type,mmax_page,mro,(int)z,(int)y,tro,(my_tm->tm_adjgbccnt-1)/(1+x-0.9*my_tm->tm_adjgbccnt),r);
+  if (r<=0.95) {
+    my_tm->tm_adjgbccnt*=mmax_page/z;
+    if (my_tm->tm_type==t_relocatable)
+      my_tm->tm_npage=z+mro;
+    else
+      my_tm->tm_maxpage=z+mro;
+    return 1;
+  }
+  return 0;
+
+}
 
 object
 alloc_object(enum type t)
@@ -221,6 +331,7 @@ alloc_object(enum type t)
 	 object obj;
 	 struct typemanager *tm;
 	 char *p;
+	 int must_have_more_pages;
 
 ONCE_MORE:
 	tm = tm_of(t);
@@ -249,25 +360,32 @@ ONCE_MORE:
 	obj->d.t = (short)t;
 	obj->d.m = FALSE;
 	return(obj);
+#ifdef SGC
 #define TOTAL_THIS_TYPE(tm) \
 	(tm->tm_nppage * (sgc_enabled ? sgc_count_type(tm->tm_type) : tm->tm_npage))
+#else
+#define TOTAL_THIS_TYPE(tm) \
+	(tm->tm_nppage * tm->tm_npage)
+#endif
 CALL_GBC:
 	GBC(tm->tm_type);
-	if (tm->tm_nfree == 0 ||
-	    ((float)tm->tm_nfree)  <   (PERCENT_FREE(tm) * TOTAL_THIS_TYPE(tm)))
-		goto EXHAUSTED;
+	must_have_more_pages = (tm->tm_nfree == 0 ||
+			  ((float)tm->tm_nfree)  <   (PERCENT_FREE(tm) * TOTAL_THIS_TYPE(tm))) ? 1 : 0;
+	if (must_have_more_pages && !IGNORE_MAX_PAGES)
+	  goto EXHAUSTED;
+	if (IGNORE_MAX_PAGES) {
+	  if ((!OPTIMIZE_MAX_PAGES || !opt_maxpage(tm)) && must_have_more_pages) {
+	     int j;
+	     
+	     tm->tm_maxpage=grow_linear((j=tm->tm_maxpage),tm->tm_growth_percent,
+					tm->tm_min_grow,tm->tm_max_grow);
+	     tm->tm_adjgbccnt*=(double)j/tm->tm_maxpage;
+	   } 
+	}
 	call_after_gbc_hook(t);
 	goto ONCE_MORE;
 
 EXHAUSTED:
-	if (IGNORE_MAX_PAGES) {
-		if (tm->tm_maxpage/2 <= 0)
-			tm->tm_maxpage += 1;
-		else
-			tm->tm_maxpage += tm->tm_maxpage/2;
-		call_after_gbc_hook(t);
-		goto ONCE_MORE;
-	}
 	GBC_enable = FALSE;
 	vs_push(type_name(t));
 	vs_push(make_fixnum(tm->tm_npage));
@@ -283,32 +401,13 @@ Use ALLOCATE to expand the space.",
 	goto ONCE_MORE;
 }
 
-static int
-grow_linear(int old, int fract, int grow_min, int grow_max) {
-  
-  int delt;
-  if (fract==0) 
-    fract=50;
-  if (grow_min==0) 
-    grow_min=1;
-  if (grow_max==0) 
-    grow_max=1000;
-
-  delt=(old*fract)/100;
-  delt= (delt < grow_min ? grow_min:
-	 delt > grow_max ? grow_max:
-	 delt);
-
-  return old + delt;
-
-}
-
 object
 make_cons(object a, object d)
 {
 	 object obj;
 	 char *p;
-	struct typemanager *tm=(&tm_table[(int)t_cons]);
+	 struct typemanager *tm=(&tm_table[(int)t_cons]);
+	 int must_have_more_pages;
 
 ONCE_MORE:
         CHECK_INTERRUPT;
@@ -339,20 +438,23 @@ ONCE_MORE:
 
 CALL_GBC:
 	GBC(t_cons);
-	if (tm->tm_nfree == 0 ||
-	    (float)tm->tm_nfree   <  PERCENT_FREE(tm) * TOTAL_THIS_TYPE(tm))
-		goto EXHAUSTED;
+	must_have_more_pages=(tm->tm_nfree == 0 ||
+			     (float)tm->tm_nfree   <  PERCENT_FREE(tm) * TOTAL_THIS_TYPE(tm)) ? 1 : 0;
+	if (must_have_more_pages && !IGNORE_MAX_PAGES)
+	  goto EXHAUSTED;
+	if (IGNORE_MAX_PAGES) {
+	  if ((!OPTIMIZE_MAX_PAGES || !opt_maxpage(tm)) && must_have_more_pages) {
+	     int j;
+	     
+	     tm->tm_maxpage=grow_linear((j=tm->tm_maxpage),tm->tm_growth_percent,
+					tm->tm_min_grow,tm->tm_max_grow);
+	     tm->tm_adjgbccnt*=(double)j/tm->tm_maxpage;
+	   } 
+	}
 	call_after_gbc_hook(t_cons);
 	goto ONCE_MORE;
 
 EXHAUSTED:
-	if (IGNORE_MAX_PAGES) {
-	  tm->tm_maxpage =
-	    grow_linear(tm->tm_maxpage,tm->tm_growth_percent,
-			tm->tm_min_grow,tm->tm_max_grow);
-	  call_after_gbc_hook(t_cons);
-	  goto ONCE_MORE;
-	}
 	GBC_enable = FALSE;
 	vs_push(make_fixnum(tm->tm_npage));
 	GBC_enable = TRUE;
@@ -417,20 +519,29 @@ void *
 alloc_contblock(size_t n) {
 	 char *p;
 	 struct contblock **cbpp;
-	 int i;
-	 int m;
+	 long i;
+	 long m;
 	 bool g;
 
-
 /*
 	printf("allocating %d-byte contiguous block...\n", n);
 */
+	 /* SGC cont pages: contiguous pointers must be aligned at
+	    CPTR_ALIGN, no smaller than sizeof (struct contblock).
+	    Here we allocate a bigger block, and rely on the fact that
+	    allocate_page returns pointers appropriately aligned,
+	    being also aligned on page boundaries.  Protection against
+	    a too small contblock was aforded before by a minimum
+	    contblock size enforced by CBMINSIZE in insert_contblock.
+	    However, this leads to a leak when many small cont blocks
+	    are allocated, e.g. with bignums, so is now removed.  CM
+	    20030827 */
 
 	g = FALSE;
-	n = ROUND_UP_PTR(n);
+	n = ROUND_UP_PTR_CONT(n);
 
 ONCE_MORE:
-	 CHECK_INTERRUPT;
+	CHECK_INTERRUPT;
 	for(cbpp= &cb_pointer; (*cbpp)!=NULL; cbpp= &(*cbpp)->cb_link)
 		if ((*cbpp)->cb_size >= n) {
 			p = (char *)(*cbpp);
@@ -443,22 +554,25 @@ ONCE_MORE:
 	m = (n + PAGESIZE - 1)/PAGESIZE;
        if(sSAignore_maximum_pagesA) {
 	if (ncbpage + m > maxcbpage || available_pages < m) {
+ 	        long j=maxcbpage;
 		if (available_pages < m)
-			sSAignore_maximum_pagesA->s.s_dbind = Cnil;
-		if (!g) {
-			GBC(t_contiguous);
-			g = TRUE;
-			call_after_gbc_hook(t_contiguous);
-			goto ONCE_MORE;
-		}
-		if (IGNORE_MAX_PAGES)
-		  {struct typemanager *tm = &tm_table[(int)t_contiguous];
-		   maxcbpage=grow_linear(maxcbpage,tm->tm_growth_percent,
-			      tm->tm_min_grow, tm->tm_max_grow);
-			g = FALSE;
-			call_after_gbc_hook(t_contiguous);
-			goto ONCE_MORE;
-		}
+		  sSAignore_maximum_pagesA->s.s_dbind = Cnil;
+		if (!g) 
+		  GBC(t_contiguous);
+		if (g && !IGNORE_MAX_PAGES)
+		  goto EXHAUSTED;
+		if (IGNORE_MAX_PAGES) {
+		  struct typemanager *tm = &tm_table[(int)t_contiguous];
+		  if ((!OPTIMIZE_MAX_PAGES || !opt_maxpage(tm)) && g) {
+		    maxcbpage=grow_linear(maxcbpage,tm->tm_growth_percent,
+					  tm->tm_min_grow, tm->tm_max_grow);
+		    tm->tm_adjgbccnt*=(double)j/maxcbpage;
+		  }
+		} 
+		g=j==maxcbpage ? TRUE : FALSE;
+		call_after_gbc_hook(t_contiguous);
+		goto ONCE_MORE;
+EXHAUSTED:
 		vs_push(make_fixnum(ncbpage));
 		CEerror("Contiguous blocks exhausted.~%\
 Currently, ~D pages are allocated.~%\
@@ -470,33 +584,93 @@ Use ALLOCATE-CONTIGUOUS-PAGES to expand the space.",
 		goto ONCE_MORE;
 	}
       }
-	p = alloc_page(m);
 
-	for (i = 0;  i < m;  i++)
+	p = alloc_page(m);
+	/* alloc_page guaranteed to return p in range for type_map references below. CM */
+	for (i = 0;  i < m;  i++) {
 		type_map[page(p) + i] = (char)t_contiguous;
+
+		/* SGC cont pages: Before this point, GCL never marked contiguous
+		   pages for SGC, causing no contiguous pages to be
+		   swept when SGC was on.  Here we follow the behavior
+		   for other pages in add_to_freelist. CM 20030827  */
+#ifdef SGC
+		if (SGC_CONT_ENABLED)
+		  sgc_type_map[page(p)+i]|= SGC_PAGE_FLAG;
+#endif
+	}
 	ncbpage += m;
 	insert_contblock(p+n, PAGESIZE*m - n);
 	return(p);
 }
 
+/* SGC cont pages: explicit free calls can come at any time, and we
+   must make sure to add the newly deallocated block to the right
+   list.  CM 20030827*/
+#ifdef SGC
+void
+insert_maybe_sgc_contblock(char *p,int s) {
+
+  struct contblock *tmp_cb_pointer;
+
+  if (SGC_CONT_ENABLED && ! SGC_PAGE_P(page(p))) {
+    tmp_cb_pointer=cb_pointer;
+    cb_pointer=old_cb_pointer;
+    sgc_enabled=0;
+    insert_contblock(p,s);
+    sgc_enabled=1;
+    old_cb_pointer=cb_pointer;
+    cb_pointer=tmp_cb_pointer;
+  } else
+    insert_contblock(p,s);
+
+}
+#endif
+
+#ifdef SGC_CONT_DEBUG
+extern void overlap_check(struct contblock *,struct contblock *);
+#endif
+
 void
 insert_contblock(char *p, int s) {
 
   struct contblock **cbpp, *cbp;
   
-  if (s < CBMINSIZE)
+  /* SGC cont pages: This used to return when s<CBMINSIZE, but we need
+     to be able to sweep small (e.g. bignum) contblocks.  FIXME:
+     should never be called with s<=0 to begin with.  CM 20030827*/
+  if (s<=0)
     return;
   ncb++;
   cbp = (struct contblock *)p;
-  cbp->cb_size = s;
+  /* SGC cont pages: allocated sizes may not be zero mod CPTR_SIZE,
+     e.g. string fillp, but alloc_contblock rounded up the allocation
+     like this, which we follow here.  CM 20030827 */
+  cbp->cb_size = ROUND_UP_PTR_CONT(s);
   for (cbpp = &cb_pointer;  *cbpp;  cbpp = &((*cbpp)->cb_link))
     if ((*cbpp)->cb_size >= s) {
+#ifdef SGC_CONT_DEBUG
+      if (*cbpp==cbp) {
+	fprintf(stderr,"Trying to install a circle at %p\n",cbp);
+	exit(1);
+      }
+      if (sgc_enabled) 
+	overlap_check(old_cb_pointer,cb_pointer);
+#endif
       cbp->cb_link = *cbpp;
       *cbpp = cbp;
+#ifdef SGC_CONT_DEBUG
+      if (sgc_enabled) 
+	overlap_check(old_cb_pointer,cb_pointer);
+#endif
       return;
     }
   cbp->cb_link = NULL;
   *cbpp = cbp;
+#ifdef SGC_CONT_DEBUG
+  if (sgc_enabled) 
+    overlap_check(old_cb_pointer,cb_pointer);
+#endif
 
 }
 
@@ -505,8 +679,8 @@ alloc_relblock(size_t n) {
 
 	 char *p;
 	 bool g;
-
-	int i;
+	 long i;
+	 int must_have_more_pages;
 
 /*
 	printf("allocating %d-byte relocatable block...\n", n);
@@ -519,68 +693,78 @@ ONCE_MORE:
         CHECK_INTERRUPT;
 
 	if (rb_limit - rb_pointer < n) {
-		if (!g && in_signal_handler == 0) {
-			GBC(t_relocatable);
-			g = TRUE;
-			{ float f1 = (float)(rb_limit - rb_pointer),
-				f2 = (float)(rb_limit - rb_start);
-
-				if ((float)f1 < PERCENT_FREE(tm_of(t_relocatable)) * f2) 
-				;
-			else
-			  {	call_after_gbc_hook(t_relocatable);
-				goto ONCE_MORE;
-					      }}
-		}
-		if (IGNORE_MAX_PAGES)
-		  {struct typemanager *tm = &tm_table[(int)t_relocatable];
-		   nrbpage=grow_linear(i=nrbpage,tm->tm_growth_percent,
-			      tm->tm_min_grow, tm->tm_max_grow);
-		   if (available_pages < 0)
-		     nrbpage = i;
-		   else {
-			  rb_end +=  (PAGESIZE* (nrbpage -i));
-			  rb_limit = rb_end - 2*RB_GETA;
-			  if (page(rb_end) - page(heap_end) !=
-			      holepage + nrbpage)
-			    FEerror("bad rb_end",0);
-			  alloc_page(-( nrbpage + holepage));
-			  g = FALSE;
-			  call_after_gbc_hook(t_relocatable);
-			  goto ONCE_MORE;
-			}
-		}
-		if (rb_limit > rb_end - 2*RB_GETA)
-			error("relocatable blocks exhausted");
-		rb_limit += RB_GETA;
-		vs_push(make_fixnum(nrbpage));
-		CEerror("Relocatable blocks exhausted.~%\
+	  if (!g && in_signal_handler == 0) 
+	    GBC(t_relocatable);
+	  must_have_more_pages=((float)(rb_limit - rb_pointer) < 
+				PERCENT_FREE(tm_of(t_relocatable)) * (float)(rb_limit - rb_start)) ? 1 : 0;
+	  if ((must_have_more_pages || g) && !IGNORE_MAX_PAGES)
+	    goto EXHAUSTED;
+	  i=nrbpage;
+	  if (IGNORE_MAX_PAGES) {
+	    struct typemanager *tm = &tm_table[(int)t_relocatable];
+	    if ((!OPTIMIZE_MAX_PAGES  || !opt_maxpage(tm)) && (g || must_have_more_pages)) {
+	      nrbpage=grow_linear(nrbpage,tm->tm_growth_percent,
+				  tm->tm_min_grow, tm->tm_max_grow);
+	      tm->tm_adjgbccnt*=(double)i/nrbpage;
+	    }
+	    if (available_pages < 0) {
+	      nrbpage = i;
+	      goto EXHAUSTED;
+	    } else if (nrbpage != i) {
+	      rb_end +=  (PAGESIZE* (nrbpage -i));
+	      rb_limit = rb_end - 2*RB_GETA;
+	      if (page(rb_end) - page(heap_end) !=
+		  holepage + nrbpage)
+		FEerror("bad rb_end",0);
+	      alloc_page(-( nrbpage + holepage));
+	    }
+	  } 
+	  g=i==nrbpage ? TRUE : FALSE;
+	  call_after_gbc_hook(t_relocatable);
+	  goto ONCE_MORE;
+EXHAUSTED:		
+	  if (rb_limit > rb_end - 2*RB_GETA)
+	    error("relocatable blocks exhausted");
+	  rb_limit += RB_GETA;
+	  vs_push(make_fixnum(nrbpage));
+	  CEerror("Relocatable blocks exhausted.~%\
 Currently, ~D pages are allocated.~%\
 Use ALLOCATE-RELOCATABLE-PAGES to expand the space.",
-			"Continues execution.", 1, vs_head,Cnil,Cnil,Cnil);
-		vs_popp;
-		g = FALSE;
-		call_after_gbc_hook(t_relocatable);
-		goto ONCE_MORE;
+		  "Continues execution.", 1, vs_head,Cnil,Cnil,Cnil);
+	  vs_popp;
+	  g = FALSE;
+	  call_after_gbc_hook(t_relocatable);
+	  goto ONCE_MORE;
 	}
 	p = rb_pointer;
 	rb_pointer += n;
 	return(p);
 }
 
+/* Add a tm_distinct field to prevent page type sharing if desired.
+   Not used now, as its never desirable from an efficiency point of
+   view, and as the only known place one must separate is cons and
+   fixnum, which are of different sizes unless PTR_ALIGN is set too
+   high (e.g. 16 on a 32bit machine).  See the ordering of init_tm
+   calls for these types below -- reversing would wind up merging the
+   types with the current algorithm.  CM 20030827 */
+
 static void
-init_tm(enum type t, char *name, int elsize, int nelts, int sgc) {
+init_tm(enum type t, char *name, int elsize, int nelts, int sgc,int distinct) {
 
   int i, j;
   int maxpage;
   /* round up to next number of pages */
   maxpage = (((nelts * elsize) + PAGESIZE -1)/PAGESIZE);
   tm_table[(int)t].tm_name = name;
-  for (j = -1, i = 0;  i < (int)t_end;  i++)
-    if (tm_table[i].tm_size != 0 &&
-	tm_table[i].tm_size >= elsize &&
-	(j < 0 || tm_table[j].tm_size > tm_table[i].tm_size))
-      j = i;
+  j=-1;
+  if (!distinct)
+    for (i = 0;  i < (int)t_end;  i++)
+      if (tm_table[i].tm_size != 0 &&
+	  tm_table[i].tm_size >= elsize &&
+	  !tm_table[i].tm_distinct &&
+	  (j < 0 || tm_table[j].tm_size > tm_table[i].tm_size))
+	j = i;
   if (j >= 0) {
     tm_table[(int)t].tm_type = (enum type)j;
     tm_table[j].tm_maxpage += maxpage;
@@ -598,6 +782,8 @@ init_tm(enum type t, char *name, int elsize, int nelts, int sgc) {
   /*tm_table[(int)t].tm_npage = 0; */  /* dont zero nrbpage.. */
   tm_table[(int)t].tm_maxpage = maxpage;
   tm_table[(int)t].tm_gbccount = 0;
+  tm_table[(int)t].tm_adjgbccnt = 0;
+  tm_table[(int)t].tm_distinct=distinct;
 #ifdef SGC	
   tm_table[(int)t].tm_sgc = sgc;
   tm_table[(int)t].tm_sgc_max = 3000;
@@ -615,9 +801,13 @@ set_maxpage(void) {
 #ifdef SGC
   page_multiple=getpagesize()/PAGESIZE;
   if (page_multiple==0) error("PAGESIZE must be factor of getpagesize()");
-  if (sgc_enabled)
-    {memory_protect(1);}
+  if (heap_end) {
+    extern long maxpage;
+    maxpage=page(heap_end);
+    memory_protect(sgc_enabled ? 1 : 0);
+  }
   if (~(-MAXPAGE) != MAXPAGE-1) error("MAXPAGE must be power of 2");
+  /* FIXME ensure core_end in range for type_map reference below.  CM */
   if (core_end)
     bzero(&sgc_type_map[ page(core_end)],MAXPAGE- page(core_end));
 #else
@@ -629,18 +819,38 @@ set_maxpage(void) {
 }
 
 
-
-
 void
-init_alloc(void) {
+gcl_init_alloc(void) {
 
-  int i;
+  long i;
   static int initialized;
+#ifdef GCL_GPROF
+   extern void *_start;
+   unsigned textpage=2*((void *)&etext-(void *)&_start)/PAGESIZE;
+#endif
   
   if (initialized) return;
   initialized=1;
   
   
+#ifdef BSD
+#ifdef RLIMIT_STACK
+ {
+
+   struct rlimit rl;
+
+   getrlimit(RLIMIT_DATA, &rl);
+   if (rl.rlim_cur != RLIM_INFINITY &&
+       (rl.rlim_max == RLIM_INFINITY ||
+	rl.rlim_max > rl.rlim_cur)) {
+     rl.rlim_cur = rl.rlim_max;
+    setrlimit(RLIMIT_DATA, &rl);
+   }
+
+ }
+#endif	
+#endif
+
 #ifndef DONT_NEED_MALLOC	
   
   {
@@ -651,6 +861,11 @@ init_alloc(void) {
 #endif	
   
   holepage = INIT_HOLEPAGE;
+#ifdef GCL_GPROF
+  if (holepage<textpage)
+     holepage=textpage;
+#endif
+
   new_holepage = HOLEPAGE;
   nrbpage = INIT_NRBPAGE;
   
@@ -688,40 +903,46 @@ init_alloc(void) {
   for (i = 0;  i < MAXPAGE;  i++)
     type_map[i] = (char)t_other;
   
+  /* Unused (at present) tm_distinct flag added.  Note that if cons
+     and fixnum share page types, errors will be introduced.
+
+     Gave each page type at least some sgc pages by default.  Of
+     course changeable by allocate-sgc.  CM 20030827 */
+
   init_tm(t_fixnum, "NFIXNUM",
-	  sizeof(struct fixnum_struct), 8192,20);
-  init_tm(t_cons, ".CONS", sizeof(struct cons), 65536 ,50 );
-  init_tm(t_structure, "SSTRUCTURE", sizeof(struct structure), 5461,0 );
-  init_tm(t_cfun, "fCFUN", sizeof(struct cfun), 4096,0  );
-  init_tm(t_sfun, "gSFUN", sizeof(struct sfun),409,0 );
-  init_tm(t_string, "\"STRING", sizeof(struct string), 5461,1  );
-  init_tm(t_array, "aARRAY", sizeof(struct array), 4681,1 );
-  init_tm(t_symbol, "|SYMBOL", sizeof(struct symbol), 3640,1 );
-  init_tm(t_bignum, "BBIGNUM", sizeof(struct bignum), 2730,0 );
-  init_tm(t_ratio, "RRATIONAL", sizeof(struct ratio), 170,0 );
+	  sizeof(struct fixnum_struct), 8192,20,0);
+  init_tm(t_cons, ".CONS", sizeof(struct cons), 65536 ,50,0 );
+  init_tm(t_structure, "SSTRUCTURE", sizeof(struct structure), 5461,1,0 );
+  init_tm(t_cfun, "fCFUN", sizeof(struct cfun), 4096,1,0  );
+  init_tm(t_sfun, "gSFUN", sizeof(struct sfun),409,1,0 );
+  init_tm(t_string, "\"STRING", sizeof(struct string), 5461,1,0  );
+  init_tm(t_array, "aARRAY", sizeof(struct array), 4681,1,0 );
+  init_tm(t_symbol, "|SYMBOL", sizeof(struct symbol), 3640,1,0 );
+  init_tm(t_bignum, "BBIGNUM", sizeof(struct bignum), 2730,1,0 );
+  init_tm(t_ratio, "RRATIONAL", sizeof(struct ratio), 170,1,0 );
   init_tm(t_shortfloat, "FSHORT-FLOAT",
-	  sizeof(struct shortfloat_struct), 256 ,1);
+	  sizeof(struct shortfloat_struct), 256 ,1,0);
   init_tm(t_longfloat, "LLONG-FLOAT",
-	  sizeof(struct longfloat_struct), 170 ,0);
-  init_tm(t_complex, "CCOMPLEX", sizeof(struct complex), 170 ,0);
-  init_tm(t_character,"#CHARACTER",sizeof(struct character), 256 ,0);
-  init_tm(t_package, ":PACKAGE", sizeof(struct package), 2*PAGESIZE / sizeof(struct package),0);
-  init_tm(t_hashtable, "hHASH-TABLE", sizeof(struct hashtable), 78,0 );
-  init_tm(t_vector, "vVECTOR", sizeof(struct vector), 146 ,0);
-  init_tm(t_bitvector, "bBIT-VECTOR", sizeof(struct bitvector), 73 ,0);
-  init_tm(t_stream, "sSTREAM", sizeof(struct stream), 78 ,0);
-  init_tm(t_random, "$RANDOM-STATE", sizeof(struct random), 256 ,0);
-  init_tm(t_readtable, "rREADTABLE", sizeof(struct readtable), 256 ,0);
-  init_tm(t_pathname, "pPATHNAME", sizeof(struct pathname), 73 ,0);
-  init_tm(t_cclosure, "cCCLOSURE", sizeof(struct cclosure), 85 ,0);
-  init_tm(t_closure, "cCLOSURE", sizeof(struct cclosure), 85 ,0);
-  init_tm(t_vfun, "VVFUN", sizeof(struct vfun), 102 ,0);
-  init_tm(t_gfun, "gGFUN", sizeof(struct sfun), 0 ,0);
-  init_tm(t_afun, "AAFUN", sizeof(struct sfun), 0 ,0);
-  init_tm(t_cfdata, "cCFDATA", sizeof(struct cfdata), 102 ,0);
-  init_tm(t_spice, "!SPICE", sizeof(struct spice), 4096 ,0);
-  init_tm(t_relocatable, "%RELOCATABLE-BLOCKS", 1000,0,20);
-  init_tm(t_contiguous, "_CONTIGUOUS-BLOCKS", 1001,0,20);
+	  sizeof(struct longfloat_struct), 170 ,1,0);
+  init_tm(t_complex, "CCOMPLEX", sizeof(struct complex), 170 ,1,0);
+  init_tm(t_character,"#CHARACTER",sizeof(struct character), 256 ,1,0);
+  init_tm(t_package, ":PACKAGE", sizeof(struct package), 2*PAGESIZE / sizeof(struct package),1,0);
+  init_tm(t_hashtable, "hHASH-TABLE", sizeof(struct hashtable), 78,1,0 );
+  init_tm(t_vector, "vVECTOR", sizeof(struct vector), 146 ,1,0);
+  init_tm(t_bitvector, "bBIT-VECTOR", sizeof(struct bitvector), 73 ,1,0);
+  init_tm(t_stream, "sSTREAM", sizeof(struct stream), 78 ,1,0);
+  init_tm(t_random, "$RANDOM-STATE", sizeof(struct random), 256 ,1,0);
+  init_tm(t_readtable, "rREADTABLE", sizeof(struct readtable), 256 ,1,0);
+  init_tm(t_pathname, "pPATHNAME", sizeof(struct pathname), 73 ,1,0);
+  init_tm(t_cclosure, "cCCLOSURE", sizeof(struct cclosure), 85 ,1,0);
+  init_tm(t_closure, "cCLOSURE", sizeof(struct cclosure), 85 ,1,0);
+  init_tm(t_vfun, "VVFUN", sizeof(struct vfun), 102 ,1,0);
+  init_tm(t_gfun, "gGFUN", sizeof(struct sfun), 0 ,1,0);
+  init_tm(t_afun, "AAFUN", sizeof(struct sfun), 0 ,1,0);
+  init_tm(t_cfdata, "cCFDATA", sizeof(struct cfdata), 102 ,1,0);
+  init_tm(t_spice, "!SPICE", sizeof(struct spice), 4096 ,1,0);
+  init_tm(t_relocatable, "%RELOCATABLE-BLOCKS", 1000,0,20,0);
+  init_tm(t_contiguous, "_CONTIGUOUS-BLOCKS", 1001,0,20,0);
   tm_table[t_relocatable].tm_nppage = PAGESIZE;
   tm_table[t_contiguous].tm_nppage = PAGESIZE;
   
@@ -729,6 +950,10 @@ init_alloc(void) {
   ncb = 0;
   ncbpage = 0;
   maxcbpage = 512;
+#ifdef GCL_GPROF
+  if (maxcbpage<textpage)
+     maxcbpage=textpage;
+#endif
   
 }
 
@@ -741,59 +966,6 @@ DEFUN_NEW("STATICP",object,fSstaticp,SI,1,1,NONE,OO,OO,OO,OO,(object x),"Tell if
 /*   FEerror("Can't get a type.", 0); */
 /* } */
 
-DEFUN_NEW("ALLOCATE",object,fSallocate,SI
-       ,2,3,NONE,OO,IO,OO,OO,(object type,fixnum npages,...),"")
-{
-
-  int nargs=VFUN_NARGS;
-  object really_do;
-  va_list ap;
-  struct typemanager *tm;
-  char *pp=NULL;
-  int t;
-  
-  really_do=Cnil;
-  if (nargs>=3) {
-    va_start(ap,npages);
-    really_do=va_arg(ap,object);
-    va_end(ap);
-  }
-  
-  CHECK_ARG_RANGE(2,3);
-  t= t_from_type(type);
-  if  (npages <= 0)
-    FEerror("Allocate takes positive argument.", 1,
-			make_fixnum(npages));
-  tm = tm_of(t);
-  if (tm->tm_npage > npages) {npages=tm->tm_npage;}
-  tm->tm_maxpage = npages;
-  if (really_do != Cnil &&
-      tm->tm_maxpage > tm->tm_npage)
-    goto ALLOCATE;
-
-  RETURN1(Ct);
-  
-       ALLOCATE:
-  if (t == t_contiguous) 
-    FUNCALL(2,fSallocate_contiguous_pages(npages,really_do));
-  
-  else
-    if (t==t_relocatable) 
-      FUNCALL(2,fSallocate_relocatable_pages(npages,really_do));
-    else {
-      
-      if (available_pages < tm->tm_maxpage - tm->tm_npage ||
-	  (pp = alloc_page(tm->tm_maxpage - tm->tm_npage)) == NULL) {
-	FEerror("Can't allocate ~D pages for ~A.", 2,
-		make_fixnum(npages), (make_simple_string(tm->tm_name+1)));
-      }
-      for (;  tm->tm_npage < tm->tm_maxpage;  pp += PAGESIZE)
-	add_page_to_freelist(pp,tm);}
-  
-  RETURN1(Ct);
-
-}
-
 static int
 t_from_type(object type) {
  
@@ -868,7 +1040,8 @@ DEFUN_NEW("ALLOCATE-CONTIGUOUS-PAGES",object,fSallocate_contiguous_pages,SI
        ,1,2,NONE,OI,OO,OO,OO,(fixnum npages,...),"")
 {
 
-  int nargs=VFUN_NARGS,i,m;
+  int nargs=VFUN_NARGS;
+  long i,m;
   object really_do;
   va_list ap;
   char *p=NULL;
@@ -884,7 +1057,7 @@ DEFUN_NEW("ALLOCATE-CONTIGUOUS-PAGES",object,fSallocate_contiguous_pages,SI
   if  (npages  < 0)
     FEerror("Allocate requires positive argument.", 0);
   if (ncbpage > npages) { 
-    printf("Allocate contiguous %ld: %d already there pages",npages,ncbpage);
+/*     printf("Allocate contiguous %ld: %d already there pages",npages,ncbpage); */
     npages=ncbpage;
   }
   maxcbpage = npages;
@@ -894,9 +1067,19 @@ DEFUN_NEW("ALLOCATE-CONTIGUOUS-PAGES",object,fSallocate_contiguous_pages,SI
   if (available_pages < m || (p = alloc_page(m)) == NULL)
     FEerror("Can't allocate ~D pages for contiguous blocks.",
 	    1, make_fixnum(npages));
-
-  for (i = 0;  i < m;  i++)
-    type_map[page(p + PAGESIZE*i)] = (char)t_contiguous;
+  
+  /* alloc_page call above guaranteed to return p in range for type_map reference below. CM */
+  for (i = 0;  i < m;  i++) {
+    type_map[page(p)+i] = (char)t_contiguous;
+    /* SGC cont pages: Before this point, GCL never marked contiguous
+       pages for SGC, causing no contiguous pages to be
+       swept when SGC was on.  Here we follow the behavior
+       for other pages in add_to_freelist. CM 20030827  */
+#ifdef SGC
+    if (SGC_CONT_ENABLED)
+      sgc_type_map[page(p)+i]|= SGC_PAGE_FLAG;
+#endif
+  }
 
   ncbpage += m;
   insert_contblock(p, PAGESIZE*m);
@@ -952,6 +1135,59 @@ DEFUN_NEW("ALLOCATE-RELOCATABLE-PAGES",object,fSallocate_relocatable_pages,SI
 
 }
 
+DEFUN_NEW("ALLOCATE",object,fSallocate,SI
+       ,2,3,NONE,OO,IO,OO,OO,(object type,fixnum npages,...),"")
+{
+
+  int nargs=VFUN_NARGS;
+  object really_do;
+  va_list ap;
+  struct typemanager *tm;
+  char *pp=NULL;
+  int t;
+  
+  really_do=Cnil;
+  if (nargs>=3) {
+    va_start(ap,npages);
+    really_do=va_arg(ap,object);
+    va_end(ap);
+  }
+  
+  CHECK_ARG_RANGE(2,3);
+  t= t_from_type(type);
+  if  (npages <= 0)
+    FEerror("Allocate takes positive argument.", 1,
+			make_fixnum(npages));
+  tm = tm_of(t);
+  if (tm->tm_npage > npages) {npages=tm->tm_npage;}
+  tm->tm_maxpage = npages;
+  if (really_do != Cnil &&
+      tm->tm_maxpage > tm->tm_npage)
+    goto ALLOCATE;
+
+  RETURN1(Ct);
+  
+       ALLOCATE:
+  if (t == t_contiguous) 
+    FUNCALL(2,FFN(fSallocate_contiguous_pages)(npages,really_do));
+  
+  else
+    if (t==t_relocatable) 
+      FUNCALL(2,FFN(fSallocate_relocatable_pages)(npages,really_do));
+    else {
+      
+      if (available_pages < tm->tm_maxpage - tm->tm_npage ||
+	  (pp = alloc_page(tm->tm_maxpage - tm->tm_npage)) == NULL) {
+	FEerror("Can't allocate ~D pages for ~A.", 2,
+		make_fixnum(npages), (make_simple_string(tm->tm_name+1)));
+      }
+      for (;  tm->tm_npage < tm->tm_maxpage;  pp += PAGESIZE)
+	add_page_to_freelist(pp,tm);}
+  
+  RETURN1(Ct);
+
+}
+
 DEFUN_NEW("ALLOCATED-RELOCATABLE-PAGES",object,fSallocated_relocatable_pages,SI
        ,0,0,NONE,OO,OO,OO,OO,(void),"")
 {
@@ -966,6 +1202,74 @@ DEFUN_NEW("GET-HOLE-SIZE",object,fSget_hole_size,SI
 	RETURN1((make_fixnum(new_holepage)));
 }
 
+
+#ifdef GCL_GPROF
+
+static unsigned long start,end,gprof_on;
+
+DEFUN_NEW("GPROF-START",object,fSgprof_start,SI
+       ,0,0,NONE,OO,OO,OO,OO,(void),"")
+{
+  extern void monstartup(unsigned long,unsigned long);
+  extern void *_start;
+
+  if (!gprof_on) {
+    start=start ? start : (unsigned long)&_start;
+    end=end ? end : (unsigned long)core_end;
+    monstartup(start,end);
+    gprof_on=1;
+  }
+
+  return Cnil;
+
+}
+
+DEFUN_NEW("GPROF-SET",object,fSgprof_set,SI
+       ,2,2,NONE,OI,IO,OO,OO,(fixnum dstart,fixnum dend),"")
+{
+
+  start=dstart;
+  end=dend;
+
+  return Cnil;
+
+}
+
+DEFUN_NEW("GPROF-QUIT",object,fSgprof_quit,SI
+       ,0,0,NONE,OO,OO,OO,OO,(void),"")
+{
+  extern void _mcleanup(void);
+  char b[PATH_MAX],b1[PATH_MAX];
+  FILE *pp;
+  unsigned n;
+
+  if (!gprof_on)
+    return Cnil;
+
+  if (!getwd(b))
+    FEerror("Cannot get working directory", 0);
+  if (chdir(P_tmpdir))
+    FEerror("Cannot change directory to tmpdir", 0);
+  _mcleanup();
+  if (snprintf(b1,sizeof(b1),"gprof %s",kcl_self)<=0)
+    FEerror("Cannot write gprof command line", 0);
+  if (!(pp=popen(b1,"r")))
+    FEerror("Cannot open gprof pipe", 0);
+  while ((n=fread(b1,1,sizeof(b1),pp)))
+    if (!fwrite(b1,1,n,stdout))
+      FEerror("Cannot write gprof output",0);
+  if (pclose(pp)<0)
+    FEerror("Cannot close gprof pipe", 0);
+  if (chdir(b))
+    FEerror("Cannot restore working directory", 0);
+  gprof_on=0;
+
+  return Cnil;
+
+}
+
+#endif
+
 DEFUN_NEW("SET-HOLE-SIZE",object,fSset_hole_size,SI
        ,1,2,NONE,OI,IO,OO,OO,(fixnum npages,...),"")
 {
@@ -999,7 +1303,7 @@ DEFUN_NEW("SET-HOLE-SIZE",object,fSset_hole_size,SI
 
 
 void
-init_alloc_function(void) {
+gcl_init_alloc_function(void) {
 }
 
 object malloc_list;
@@ -1086,19 +1390,26 @@ static char *baby_malloc(n)
 
 void *
 malloc(size_t size) {
-        object x;
-	
+	static int in_malloc;
+
+	if (in_malloc)
+	  return NULL;
+	in_malloc=1;
+
 /*  #ifdef HAVE_LIBBFD */
 /*  	if (in_bfd_init) */
 /*  	  return bfd_malloc(size); */
 /*  #endif */
 
 #ifdef BABY_MALLOC_SIZE
-	if (GBC_enable == 0) return baby_malloc(size);
+	if (GBC_enable == 0) {
+	  in_malloc = 0;
+	  return baby_malloc(size);
+	}
 #else	
 	if (GBC_enable==0) {
 	   if ( initflag ==0)
-	     init_alloc();
+	     gcl_init_alloc();
 	   else {
 #ifdef	     RECREATE_HEAP
 	     	RECREATE_HEAP
@@ -1109,16 +1420,17 @@ malloc(size_t size) {
 
 #endif	
       
+	malloc_list = make_cons(Cnil, malloc_list);
 
-	x = alloc_simple_string(size);
+	malloc_list->c.c_car = alloc_simple_string(size);
 
-	x->st.st_self = alloc_contblock(size);
+	malloc_list->c.c_car->st.st_self = alloc_contblock(size);
 #ifdef SGC
-	perm_writable(x->st.st_self,size);
+	perm_writable(malloc_list->c.c_car->st.st_self,size);
 #endif
-	malloc_list = make_cons(x, malloc_list);
 
-	return(x->st.st_self);
+	in_malloc=0;
+	return(malloc_list->c.c_car->st.st_self);
 }
 
 
@@ -1140,8 +1452,14 @@ free(void *ptr)
 #endif	
 	for (p = &malloc_list; *p && !endp(*p);  p = &((*p)->c.c_cdr))
 		if ((*p)->c.c_car->st.st_self == ptr) {
-			insert_contblock((*p)->c.c_car->st.st_self,
+/* SGC contblock pages: Its possible this is on an old page CM 20030827 */
+#ifdef SGC
+ 			insert_maybe_sgc_contblock((*p)->c.c_car->st.st_self,
+						   (*p)->c.c_car->st.st_dim);
+#else
+ 			insert_contblock((*p)->c.c_car->st.st_self,
 					 (*p)->c.c_car->st.st_dim);
+#endif
 			(*p)->c.c_car->st.st_self = NULL;
 			*p = (*p)->c.c_cdr;
 			return ;
@@ -1161,7 +1479,7 @@ realloc(void *ptr, size_t size) {
   int i, j;
   /* was allocated by baby_malloc */
 #ifdef BABY_MALLOC_SIZE	
-  if (ptr >= baby_malloc_data && ptr -baby_malloc_data <BABY_MALLOC_SIZE)
+  if (ptr >= (void*)baby_malloc_data && ptr - (void*)baby_malloc_data <BABY_MALLOC_SIZE)
     {
       int dim = ((int *)ptr)[-1];
       if (dim > size)
@@ -1189,7 +1507,12 @@ realloc(void *ptr, size_t size) {
 	x->st.st_fillp = x->st.st_dim = size;
 	for (i = 0;  i < size;  i++)
 	  x->st.st_self[i] = ((char *)ptr)[i];
-	insert_contblock(ptr, j);
+/* SGC contblock pages: Its possible this is on an old page CM 20030827 */
+#ifdef SGC
+ 	insert_maybe_sgc_contblock(ptr, j);
+#else
+ 	insert_contblock(ptr, j);
+#endif
 	return(x->st.st_self);
       }
     }
