@@ -120,7 +120,7 @@ ihs_top_function_name(ihs_ptr h)
 
 
 
-static object
+object
 Icall_error_handler(object error_name,object error_format_string,int nfmt_args,...)
 { object b[20];
   b[0]= error_name;
@@ -238,6 +238,7 @@ DEFUNO_NEW("CERROR",object,fLcerror,LISP
   b[3]=null_string;
   b[4]=fmt_string;
   i=4;
+  n--;
   va_start(ap,fmt_string);
   while (--n)
     b[++i]=va_arg(ap,object);
@@ -282,7 +283,7 @@ FEerror(char *s,int num,...)
     arg4=va_arg(args,object);
   va_end(args);
 
-  fLerror(make_simple_string(s),arg1,arg2,arg3,arg4);
+  FFN(fLerror)(make_simple_string(s),arg1,arg2,arg3,arg4);
   
 }
 
@@ -391,7 +392,7 @@ object
 CEerror(char *error_str, char *cont_str, int num, object arg1, object arg2, object arg3, object arg4)
 {
         VFUN_NARGS=num+2; 
-        return fLcerror(make_simple_string(cont_str),
+        return FFN(fLcerror)(make_simple_string(cont_str),
 			make_simple_string(error_str),
 			arg1,arg2,arg3,arg4);
 }
@@ -812,7 +813,7 @@ DEF_ORDINARY("CATCHALL",sKcatchall,KEYWORD,"");
 
 
 void
-init_error(void)
+gcl_init_error(void)
 {
 	make_function("ERROR", Lerror);
 	make_function("CERROR", Lcerror);
