@@ -93,12 +93,12 @@ object sSAbreak_stepA;
 static void
 quick_call_sfun(object fun)
 { DEBUG_AVMA
-  int i,n;
+  int i=fun->sfn.sfn_argd,n=SFUN_NARGS(i);
   enum ftype restype;
   object *x,res,*base;
-  object temp_ar[10];
-  i=fun->sfn.sfn_argd;
-  n=SFUN_NARGS(i);
+  object *temp_ar=alloca(n*sizeof(object));
+/*   i=fun->sfn.sfn_argd; */
+/*   n=SFUN_NARGS(i); */
   base = vs_base;
   if (n != vs_top - base)
     {check_arg_failed(n);}
@@ -1180,8 +1180,7 @@ DEFUNO_NEW("EVAL",object,fLeval,LISP
 	return Ivs_values();
 }
 
-void
-Levalhook(void)
+LFD(Levalhook)(void)
 {
 	object env;
 	bds_ptr old_bds_top = bds_top;
@@ -1211,8 +1210,7 @@ Levalhook(void)
 	bds_unwind(old_bds_top);
 }
 
-void
-Lapplyhook(void)
+LFD(Lapplyhook)(void)
 {
 
 	object env;
@@ -1524,7 +1522,7 @@ DEFVAR("*BREAK-POINTS*",sSAbreak_pointsA,SI,Cnil,"");
 DEFVAR("*BREAK-STEP*",sSAbreak_stepA,SI,Cnil,"");
 
 void
-init_eval(void)
+gcl_init_eval(void)
 {
 
 
