@@ -31,6 +31,7 @@
 #include <string.h>
 #include <stdlib.h>
 #include "include.h"
+#include "page.h"
 
 
 #ifdef SGC
@@ -66,18 +67,26 @@ mark_object(object);
 */ 
 #define BBITS_CHAR 3
 
-#if SIZEOF_LONG_P == 4
+#if SIZEOF_LONG == 4
 #define BBYTES_LONG 2
-#elif SIZEOF_LONG_P == 8
+#elif SIZEOF_LONG == 8
 #define BBYTES_LONG 3
 #else
-#error Do not recognize SIZEOF_LONG_P
+#error Do not recognize SIZEOF_LONG
+#endif
+
+#if CPTR_ALIGN == 8
+#define BBYTES_CONTBLOCK 3
+#elif CPTR_ALIGN == 16
+#define BBYTES_CONTBLOCK 4
+#else
+#error Do not recognize CPTR_ALIGN
 #endif
 
 #define BBITS_LONG (BBYTES_LONG+BBITS_CHAR)
-#define BCHARS_TABLE (BBITS_LONG+BBYTES_LONG)
+#define BCHARS_TABLE (BBITS_LONG+BBYTES_CONTBLOCK)
 
-#define Shamt(x) (((((unsigned long) x) >> BBYTES_LONG) & ~(~0 << BBITS_LONG)))
+#define Shamt(x) (((((unsigned long) x) >> BBYTES_CONTBLOCK) & ~(~0 << BBITS_LONG)))
 #define Madr(x) (mark_table+((((unsigned long) x) - ((unsigned long)DBEGIN)) >> (BCHARS_TABLE)))
 #define get_mark_bit(x) (*(Madr(x)) >> Shamt(x) & 1)
 #define set_mark_bit(x) ((*(Madr(x))) |= ((unsigned long)1 << Shamt(x)))
@@ -96,15 +105,17 @@ bool ovm_process_created;
 bool saving_system;
 static int gc_time = -1;
 static int gc_start = 0;
+static int gc_recursive = 0;
 int runtime(void);
+#ifdef SGC
 int sgc_enabled=0;
-int  first_protectable_page =0;
+#endif
+long  first_protectable_page =0;
 
 
 
 static char *copy_relblock(char *p, int s);
 
-#include "page.h"
 
 
 #ifdef MV
@@ -117,13 +128,13 @@ long real_maxpage;
 long new_holepage;
 
 #define	available_pages	\
-	(real_maxpage-page(heap_end)-new_holepage-2*nrbpage-real_maxpage/32)
+	(real_maxpage-page(heap_end)-(new_holepage>=holepage ? new_holepage : holepage)-2*nrbpage-real_maxpage/32)
 
 struct apage {
   char apage_self[PAGESIZE];
 };
 
-int maxpage;
+long maxpage;
 
 object sSAnotify_gbcA;
 
@@ -168,6 +179,7 @@ struct {
 } mark_origin_block[MARK_ORIGIN_BLOCK_MAX];
 int mark_origin_block_max;
 
+/* must be a long * to match with SIZEOF_LONG usage above*/
 long *mark_table;
 
 enum type what_to_collect;
@@ -177,10 +189,14 @@ enum type what_to_collect;
 void
 enter_mark_origin(object *p)
 {
+  unsigned long np=page(p);
+/*   if (np>=MAXPAGE) */
+/*     error("Address supplied to enter_mar_origin out of range"); */
   if (mark_origin_max >= MARK_ORIGIN_MAX)
     error("too many mark origins");
 #ifdef SGC
-  sgc_type_map[page(p)] |= SGC_PERM_WRITABLE ;
+  if (np<MAXPAGE)
+    sgc_type_map[np] |= SGC_PERM_WRITABLE ;
 #endif	
   mark_origin[mark_origin_max++] = p;
 }
@@ -410,7 +426,7 @@ mark_object(object x) {
 	  mark_contblock(cp, j);
       } else if (x->a.a_displaced == Cnil) {
 #ifdef HAVE_ALLOCA
-	if (cp <= core_end)  /* only if body of array not on C stack */
+	if (!NULL_OR_ON_C_STACK(cp))  /* only if body of array not on C stack */
 #endif			  
 	  x->a.a_self = (object *)copy_relblock(cp, j);}
       else if (x->a.a_displaced->c.c_car == Cnil) {
@@ -662,7 +678,8 @@ static long *c_stack_where;
 static void
 mark_stack_carefully(void *topv, void *bottomv, int offset) {
 
-  int p,m,pageoffset;
+  long m,pageoffset;
+  unsigned long p;
   object x;
   struct typemanager *tm;
   register long *j;
@@ -695,7 +712,7 @@ mark_stack_carefully(void *topv, void *bottomv, int offset) {
 	if (m==TRUE) continue;
 	if (m!=0) {
 	  fprintf(stdout,
-		  "**bad value %d of d.m in gbc page %d skipping mark**"
+		  "**bad value %ld of d.m in gbc page %ld skipping mark**"
 		  ,m,p);fflush(stdout);
 	  continue;
 	}
@@ -902,7 +919,7 @@ mark_c_stack(jmp_buf env1, int n, void (*fn)(void *,void *,int)) {
 static void
 sweep_phase(void) {
 
-  STATIC int i, j, k;
+  STATIC long i, j, k;
   STATIC object x;
   STATIC char *p;
   STATIC struct typemanager *tm;
@@ -992,7 +1009,7 @@ sweep_phase(void) {
 static void
 contblock_sweep_phase(void) {
 
-  STATIC int i, j;
+  STATIC long i, j;
   STATIC char *s, *e, *p, *q;
   STATIC struct contblock *cbp;
   
@@ -1011,19 +1028,24 @@ contblock_sweep_phase(void) {
     e = pagetochar(j);
     for (p = s;  p < e;) {
       if (get_mark_bit((int *)p)) {
-	p += PTR_ALIGN;
+	/* SGC cont pages: cont blocks must be no smaller than
+	   sizeof(struct contblock), and must not have a sweep
+	   granularity greater than this amount (e.g. CPTR_ALIGN) if
+	   contblock leaks are to be avoided.  Used to be aligned at
+	   PTR_ALIGN. CM 20030827 */
+	p += CPTR_ALIGN;
 	continue;
       }
-      q = p + PTR_ALIGN;
+      q = p + CPTR_ALIGN;
       while (q < e) {
 	if (!get_mark_bit((int *)q)) {
-	  q += PTR_ALIGN;
+	  q += CPTR_ALIGN;
 	  continue;
 	}
 	break;
       }
       insert_contblock(p, q - p);
-      p = q + PTR_ALIGN;
+      p = q + CPTR_ALIGN;
     }
     i = j + 1;
   }
@@ -1044,9 +1066,11 @@ char *old_rb_start;
 void
 GBC(enum type t) {
 
-  int i, j;
+  long i, j;
   struct apage *pp, *qq;
+#ifdef SGC
   int in_sgc = sgc_enabled;
+#endif
 #ifdef DEBUG
   int tm=0;
 #endif
@@ -1063,11 +1087,12 @@ GBC(enum type t) {
   
   if (saving_system)
     {t = t_contiguous; gc_time = -1;
+#ifdef SGC
     if(sgc_enabled) sgc_quit();
-    
+#endif    
     }
-  
-  
+
+
 #ifdef DEBUG
   debug = symbol_value(sSAgbc_messageA) != Cnil;
 #endif
@@ -1075,15 +1100,23 @@ GBC(enum type t) {
   what_to_collect = t;
   
   tm_table[(int)t].tm_gbccount++;
+  tm_table[(int)t].tm_adjgbccnt++;
   
 #ifdef DEBUG
   if (debug || (sSAnotify_gbcA->s.s_dbind != Cnil)) {
     
     if (gc_time < 0) gc_time=0;
-    printf("[%s for %d %s pages..",
+#ifdef SGC
+    printf("[%s for %ld %s pages..",
 	   (sgc_enabled ? "SGC" : "GC"),
 	   (sgc_enabled ? sgc_count_type(t) : tm_of(t)->tm_npage),
 	   (tm_table[(int)t].tm_name)+1);
+#else
+    printf("[%s for %d %s pages..",
+	   ("GC"),
+	   (tm_of(t)->tm_npage),
+	   (tm_table[(int)t].tm_name)+1);
+#endif
 #ifdef SGC
     if(sgc_enabled)
       printf("(%d writable)..",sgc_count_writable(page(core_end)));
@@ -1091,12 +1124,12 @@ GBC(enum type t) {
     fflush(stdout);
   }
 #endif
-  if (gc_time >=0) {gc_start=runtime();}
+  if (gc_time >=0 && !gc_recursive++) {gc_start=runtime();}
   
   maxpage = page(heap_end);
   
   if ((int)t >= (int)t_contiguous) {
-    j = maxpage*(PAGESIZE/(sizeof(int)*sizeof(int)*CHAR_SIZE)) ;
+    j = maxpage*(PAGESIZE/(CPTR_ALIGN*SIZEOF_LONG*CHAR_SIZE)) ;
     /*
       (PAGESIZE / sizeof(int)) = x * (sizeof(int)*CHAR_SIZE)
       eg if PAGESIZE = 2048  x=16
@@ -1108,11 +1141,18 @@ GBC(enum type t) {
       j = 0;
     /* if in sgc we don't need more pages below hole
        just more relocatable or cleaning it */
-    if (sgc_enabled ==0 && holepage < new_holepage)
-      holepage = new_holepage;
+#ifdef SGC
+    if (sgc_enabled==0) 
+#endif
+      if (holepage < new_holepage)
+	holepage = new_holepage;
     
+#ifdef SGC
     i = rb_pointer - (sgc_enabled ? old_rb_start : rb_start);
-    
+#else
+    i = rb_pointer - rb_start;
+#endif    
+
     if (nrbpage > (real_maxpage-page(heap_end)
 		   -holepage-real_maxpage/32)/2) {
       if (i > nrbpage*PAGESIZE)
@@ -1126,7 +1166,9 @@ GBC(enum type t) {
     if (saving_system)
       rb_start = heap_end;
     else
+#ifdef SGC
       if (sgc_enabled==0)
+#endif
 	{rb_start = heap_end + PAGESIZE*holepage;}
     
     rb_end = heap_end + (holepage + nrbpage) *PAGESIZE;
@@ -1254,7 +1296,7 @@ GBC(enum type t) {
   if (debug) {
     for (i = 0, j = 0;  i < (int)t_end;  i++) {
       if (tm_table[i].tm_type == (enum type)i) {
-	printf("%13s: %8d used %8d free %4d/%d pages\n",
+	printf("%13s: %8ld used %8ld free %4ld/%ld pages\n",
 	       tm_table[i].tm_name,
 	       TM_NUSED(tm_table[i]),
 	       tm_table[i].tm_nfree,
@@ -1266,10 +1308,10 @@ GBC(enum type t) {
 	       tm_table[i].tm_name,
 	       tm_table[(int)tm_table[i].tm_type].tm_name);
     }
-    printf("contblock: %d blocks %d pages\n", ncb, ncbpage);
-    printf("hole: %d pages\n", holepage);
-    printf("relblock: %d bytes used %d bytes free %d pages\n",
-	   (int)(rb_pointer - rb_start), (int)(rb_end - rb_pointer), nrbpage);
+    printf("contblock: %ld blocks %ld pages\n", ncb, ncbpage);
+    printf("hole: %ld pages\n", holepage);
+    printf("relblock: %ld bytes used %ld bytes free %ld pages\n",
+	   (long)(rb_pointer - rb_start), (long)(rb_end - rb_pointer), nrbpage);
     printf("GBC ended\n");
     fflush(stdout);
   }
@@ -1277,6 +1319,11 @@ GBC(enum type t) {
   
   interrupt_enable = TRUE;
   
+#ifdef SGC
+  if (in_sgc && sgc_enabled==0)
+    sgc_start();
+#endif
+  
   if (saving_system) {
     j = (rb_pointer-rb_start+PAGESIZE-1) / PAGESIZE;
     
@@ -1296,19 +1343,22 @@ GBC(enum type t) {
       }
     core_end = heap_end + PAGESIZE;
     
-    for (i = 0;  i < maxpage;  i++)
-      if ((enum type)type_map[i] == t_contiguous)
-	type_map[i] = (char)t_other;
-    cb_pointer = NULL;
-    maxcbpage -= ncbpage;
-    if (maxcbpage < 100)
-      maxcbpage = 100;
-    ncbpage = 0;
-    ncb = 0;
+/*     for (i = 0;  i < maxpage;  i++) */
+/*       if ((enum type)type_map[i] == t_contiguous) */
+/* 	type_map[i] = (char)t_other; */
+/*     cb_pointer = NULL; */
+/*     maxcbpage -= ncbpage; */
+/*     if (maxcbpage < 100) */
+/*       maxcbpage = 100; */
+/*     ncbpage = 0; */
+/*     ncb = 0; */
     
     /* hmm.... why is this test necessary.*/
-    if (sgc_enabled==0) {holepage = new_holepage;
-    nrbpage = INIT_NRBPAGE;}
+#ifdef SGC
+    if (sgc_enabled==0) 
+#endif
+      {holepage = new_holepage;
+      nrbpage = INIT_NRBPAGE;}
     
     if (nrbpage < 0)
       error("no relocatable pages left");
@@ -1322,18 +1372,16 @@ GBC(enum type t) {
   if (GBC_exit_hook != NULL)
     (*GBC_exit_hook)();
   
-  
-  if (in_sgc && sgc_enabled==0)
-    sgc_start();
-  
-  if(gc_time>=0) {gc_time=gc_time+(gc_start=(runtime()-gc_start));}
+  if(gc_time>=0 && !--gc_recursive) {gc_time=gc_time+(gc_start=(runtime()-gc_start));}
   
   if (sSAnotify_gbcA->s.s_dbind != Cnil) {
     
-    fprintf(stdout, "(T=%d).GC finished]\n",
-	    gc_start
-	    );
+    if (gc_recursive)
+      fprintf(stdout, "(T=...).GC finished]\n");
+    else
+      fprintf(stdout, "(T=%d).GC finished]\n",gc_start);
     fflush(stdout);
+
   }
   
   
@@ -1341,7 +1389,7 @@ GBC(enum type t) {
 }
 
 static void
-siLroom_report(void) {
+FFN(siLroom_report)(void) {
 
   int i;
   
@@ -1380,14 +1428,14 @@ siLroom_report(void) {
 }
 
 static void
-siLreset_gbc_count(void) {
+FFN(siLreset_gbc_count)(void) {
 
   int i;
   
   check_arg(0);
   
   for (i = 0;  i < (int)t_other;  i++)
-    tm_table[i].tm_gbccount = 0;
+    tm_table[i].tm_gbccount = tm_table[i].tm_adjgbccnt = 0;
 }
 
 /* copy S bytes starting at P to beyond rb_pointer1 (temporarily)
@@ -1415,14 +1463,17 @@ static void
 mark_contblock(void *p, int s) {
 
   STATIC char *q;
-  STATIC int *x, *y;
+  STATIC char *x, *y;
+  long np=page(p);
   
-  if (!MAYBE_DATA_P(p) || (enum type)type_map[page(p)] != t_contiguous)
+  if (!MAYBE_DATA_P(p) || np >= MAXPAGE || (enum type)type_map[page(p)] != t_contiguous)
     return;
   q = p + s;
-  x = (int *)ROUND_DOWN_PTR(p);
-  y = (int *)ROUND_UP_PTR(q);
-  for (;  x < y;  x++)
+  /* SGC cont pages: contblock pages must be no smaller than
+     sizeof(struct contblock).  CM 20030827 */
+  x = (char *)ROUND_DOWN_PTR_CONT(p);
+  y = (char *)ROUND_UP_PTR_CONT(q);
+  for (;  x < y;  x+=CPTR_ALIGN)
     set_mark_bit(x);
 }
 
@@ -1442,7 +1493,7 @@ DEFUN_NEW("GBC",object,fLgbc,LISP
 }
 
 static void
-siLgbc_time(void) {
+FFN(siLgbc_time)(void) {
   if (vs_top>vs_base)
     gc_time=fix(vs_base[0]);
   else {
@@ -1461,7 +1512,7 @@ DEFVAR("*GBC-MESSAGE*",sSAgbc_messageA,SI,Cnil,"");
 #endif
 
 void
-init_GBC(void) {
+gcl_init_GBC(void) {
 
   make_si_function("ROOM-REPORT", siLroom_report);
   make_si_function("RESET-GBC-COUNT", siLreset_gbc_count);
@@ -1471,7 +1522,7 @@ init_GBC(void) {
   /* we use that maxpage is a power of 2 in this
      case, to quickly be able to look in our table */ 
   { 
-    int i,j;
+    long i,j;
   
     for(i=j=1 ; i< 32 ; i++) 
       if (MAXPAGE == (1 <<i))
