@@ -49,8 +49,6 @@ void initialize_process();
 #include "include.h"
 #ifdef UNIX
 #include <signal.h>
-
-
 #endif
 #include "page.h"
 
@@ -104,7 +102,9 @@ unsigned int _stacktop, _stackbottom;
 
 int cssize;
 
+#ifdef SGC
 int sgc_enabled;
+#endif
 void install_segmentation_catcher(void);
 
 #ifndef SIG_STACK_SIZE
@@ -116,6 +116,12 @@ void install_segmentation_catcher(void);
 #endif
 #endif
 
+#ifdef NEED_NONRANDOM_SBRK
+#include <syscall.h>
+#include <linux/personality.h>
+#include <unistd.h>
+#endif
+
 int
 main(int argc, char **argv, char **envp) {
 #ifdef BSD
@@ -124,6 +130,20 @@ main(int argc, char **argv, char **envp) {
 #endif
 #endif
 
+#ifdef NEED_NONRANDOM_SBRK
+#if SIZEOF_LONG == 4
+	if (!syscall(SYS_personality,PER_LINUX32))
+#else
+        if (!syscall(SYS_personality,PER_LINUX))
+#endif
+	  execvp(argv[0],argv);
+#endif
+
+#if defined(DARWIN)
+    extern void init_darwin_zone_compat ();
+    init_darwin_zone_compat ();
+#endif
+
 #ifdef RECREATE_HEAP
 	RECREATE_HEAP
 #endif
@@ -141,6 +161,9 @@ main(int argc, char **argv, char **envp) {
 	setbuf(stdout, stdout_buf);
 #ifdef _WIN32
 	_fmode = _O_BINARY;
+        _setmode( _fileno( stdin ), _O_BINARY );
+        _setmode( _fileno( stdout ), _O_BINARY );
+        _setmode( _fileno( stderr ), _O_BINARY );
 #endif
 	ARGC = argc;
 	ARGV = argv;
@@ -153,7 +176,11 @@ main(int argc, char **argv, char **envp) {
 	if (argv[0][0] != '/')
 		error("can't get the program name");
 */
+#ifdef GET_FULL_PATH_SELF
+	GET_FULL_PATH_SELF(kcl_self);
+#else
 	kcl_self = argv[0];
+#endif
 #ifdef FIX_FILENAME
 	{ int n = strlen(kcl_self);
 	 FIX_FILENAME(Cnil,kcl_self);
@@ -234,15 +261,37 @@ main(int argc, char **argv, char **envp) {
 	install_segmentation_catcher();
 
 #ifdef BSD
-#ifndef MAX_STACK_SIZE
-#define MAX_STACK_SIZE (1<<23) /* 8Mb */
-#endif
 #ifdef RLIMIT_STACK
-	getrlimit(RLIMIT_STACK, &rl);
-        if (rl.rlim_cur == RLIM_INFINITY || 
-	    rl.rlim_cur > MAX_STACK_SIZE)
-	  rl.rlim_cur=MAX_STACK_SIZE;
-	cssize = rl.rlim_cur/4 - 4*CSGETA;
+	{
+	  unsigned long mss;
+	  mss=16*sizeof(short)*MAXPAGE; /* i.e. short foo[MAXPAGE] on stack in sgc_start */
+	  if (getrlimit(RLIMIT_STACK, &rl))
+	    error("Cannot get stack rlimit\n");
+	  if (rl.rlim_max != RLIM_INFINITY && rl.rlim_max < mss)
+	    mss=rl.rlim_max;
+	  if (rl.rlim_cur == RLIM_INFINITY || 
+	      rl.rlim_cur != mss) {
+	    rl.rlim_cur=mss;
+	    if (setrlimit(RLIMIT_STACK,&rl))
+	      error("Cannot set stack rlimit\n");
+	  }
+	  cssize = rl.rlim_cur/sizeof(*cs_org) - sizeof(*cs_org)*CSGETA;
+        }
+
+
+	/* Maybe the soft limit for data segment size is lower than the
+	 * hard limit.  In that case, we want as much as possible.
+	 */
+	if (getrlimit(RLIMIT_DATA, &rl))
+	    error("Cannot get data rlimit\n");
+	if (rl.rlim_cur != RLIM_INFINITY &&
+	    (rl.rlim_max == RLIM_INFINITY ||
+	     rl.rlim_max > rl.rlim_cur)) {
+	  rl.rlim_cur = rl.rlim_max;
+	  if (setrlimit(RLIMIT_DATA, &rl))
+	    error("Cannot set data rlimit\n");
+	}
+
 #endif	
 #endif
 
@@ -273,13 +322,16 @@ main(int argc, char **argv, char **envp) {
 #endif	
 #endif	
 	
+#ifdef SGC
+	memprotect_test_reset();
+#endif
 
 	if (initflag) {
 		if (saving_system) {
 			saving_system = FALSE;
 			terminal_io->sm.sm_object0->sm.sm_fp = stdin;
 			terminal_io->sm.sm_object1->sm.sm_fp = stdout;
-			init_big1();
+			gcl_init_big1();
 #ifdef INIT_CORE_END
 			INIT_CORE_END
 #endif			  
@@ -306,7 +358,7 @@ main(int argc, char **argv, char **envp) {
 		ovm_process_created = 1;
 #endif
 #ifdef HAVE_READLINE
-		init_readline_function();
+		gcl_init_readline_function();
 #endif
 	      again:
 		super_funcall(sStop_level);
@@ -337,7 +389,8 @@ main(int argc, char **argv, char **envp) {
 
 	CMPtemp = CMPtemp1 = CMPtemp2 = CMPtemp3 = OBJNULL;
 
-	init_init();
+	parse_plt();
+	gcl_init_init();
 
 	sLApackageA->s.s_dbind = user_package;
 
@@ -371,9 +424,11 @@ error(char *s)
 {
         if (catch_fatal>0 && interrupt_enable )
             {catch_fatal = -1;
+#ifdef SGC
 	   if (sgc_enabled)
 	     { sgc_quit();}
 	   if (sgc_enabled==0)
+#endif
 	     { install_segmentation_catcher() ;}
 	   FEerror("Caught fatal error [memory may be damaged]",0); }
 	printf("\nUnrecoverable error: %s.\n", s);
@@ -395,7 +450,7 @@ initlisp(void) {
 	  { /* check person has correct definition of above */
 	    error("NULL_OR_ON_C_STACK macro invalid");
 	  }
-	init_alloc();
+	gcl_init_alloc();
 
 	Dotnil_body.t = (short)t_symbol;
 	Dotnil_body.s_dbind = Dotnil;
@@ -403,8 +458,8 @@ initlisp(void) {
 	Dotnil_body.s_fillp = 6;
 	Dotnil_body.s_self = "DOTNIL";
 	Dotnil_body.s_gfdef = OBJNULL;
-	Dotnil_body.s_plist = Dotnil;
-	Dotnil_body.s_hpack = Dotnil;
+	Dotnil_body.s_plist = Cnil;
+	Dotnil_body.s_hpack = Cnil;
 	Dotnil_body.s_stype = (short)stp_constant;
 	Dotnil_body.s_mflag = FALSE;
 	
@@ -430,9 +485,9 @@ initlisp(void) {
 	Ct_body.s_stype = (short)stp_constant;
 	Ct_body.s_mflag = FALSE;
 	
-	init_symbol();
+	gcl_init_symbol();
 
-	init_package();
+	gcl_init_package();
 
 	Cnil->s.s_hpack = lisp_package;
 	import(Cnil, lisp_package);
@@ -452,8 +507,8 @@ initlisp(void) {
 	export(Ct, common_lisp_package);
 #endif
 
-	sLquote = make_ordinary("QUOTE");
-	sLfunction = make_ordinary("FUNCTION");
+/* 	sLquote = make_ordinary("QUOTE"); */
+/* 	sLfunction = make_ordinary("FUNCTION"); */
 	sLlambda = make_ordinary("LAMBDA");
 	sLlambda_block = make_ordinary("LAMBDA-BLOCK");
 	sLlambda_closure = make_ordinary("LAMBDA-CLOSURE");
@@ -462,21 +517,21 @@ initlisp(void) {
 
 	
 	NewInit();
-	init_typespec();
-	init_number();
-	init_character();
+	gcl_init_typespec();
+	gcl_init_number();
+	gcl_init_character();
 
-	init_read();
-	init_bind();
-	init_pathname();
-	init_print();
-	init_GBC();
+	gcl_init_read();
+	gcl_init_bind();
+	gcl_init_pathname();
+	gcl_init_print();
+	gcl_init_GBC();
 
 #if defined ( UNIX ) || defined ( __MINGW32__ )
 #ifndef DGUX
-	init_unixfasl();
-	init_unixsys();
-	init_unixsave();
+	gcl_init_unixfasl();
+	gcl_init_unixsys();
+	gcl_init_unixsave();
 #else
 
 
@@ -484,57 +539,58 @@ initlisp(void) {
 #endif
 #endif
 
-	init_alloc_function();
-	init_array_function();
-	init_character_function();
-	init_file_function();
-	init_list_function();
-	init_package_function();
-	init_pathname_function();
-	init_predicate_function();
-	init_print_function();
-	init_read_function();
-	init_sequence_function();
+	gcl_init_alloc_function();
+	gcl_init_array_function();
+	gcl_init_character_function();
+	gcl_init_file_function();
+	gcl_init_list_function();
+	gcl_init_package_function();
+	gcl_init_pathname_function();
+	gcl_init_predicate_function();
+	gcl_init_print_function();
+	gcl_init_read_function();
+	gcl_init_sequence_function();
 #if  defined(KCLOVM) || defined(RUN_PROCESS)
-	init_socket_function();
+	gcl_init_socket_function();
 #endif	
-	init_structure_function();
-	init_string_function();
-	init_symbol_function();
-	init_typespec_function();
-	init_hash();
-	init_cfun();
+	gcl_init_structure_function();
+	gcl_init_string_function();
+	gcl_init_symbol_function();
+	gcl_init_typespec_function();
+	gcl_init_hash();
+	gcl_init_cfun();
 
 #ifdef UNIX
-	init_unixfsys();
-	init_unixtime();
+	gcl_init_unixfsys();
+	gcl_init_unixtime();
 #endif
-	init_eval();
-	init_lex();
-	init_prog();
-	init_catch();
-	init_block();
-        init_macros();
-	init_conditional();
-	init_reference();
-	init_assignment();
-	init_multival();
-	init_error();
-	init_let();
-	init_mapfun();
-	init_iteration();
-	init_toplevel();
+	gcl_init_eval();
+	gcl_init_lex();
+	gcl_init_prog();
+	gcl_init_catch();
+	gcl_init_block();
+        gcl_init_macros();
+	gcl_init_conditional();
+	gcl_init_reference();
+	gcl_init_assignment();
+	gcl_init_multival();
+	gcl_init_error();
+	gcl_init_let();
+	gcl_init_mapfun();
+	gcl_init_iteration();
+	gcl_init_toplevel();
 
-	init_cmpaux();
+	gcl_init_cmpaux();
 
 	init_main();
 
-	init_format();
-	init_links();
+	gcl_init_format();
+	gcl_init_links();
 
-	init_fat_string();
+	gcl_init_fat_string();
+	gcl_init_sfasl();
 #ifdef CMAC
-	init_cmac();
+	gcl_init_cmac();
 #endif	
 
 }
@@ -625,11 +681,11 @@ DEFUNO_NEW("BYE",object,fLbye,LISP
 
 DEFUN_NEW("QUIT",object,fLquit,LISP
        ,0,1,NONE,OI,OO,OO,OO,(fixnum exitc),"")
-{	return fLbye(exitc); }
+{	return FFN(fLbye)(exitc); }
  
-DEFUN_NEW("EXIT",object,fLexit,LISP
-       ,0,1,NONE,OI,OO,OO,OO,(fixnum exitc),"")
-{	return fLbye(exitc); }
+/* DEFUN_NEW("EXIT",object,fLexit,LISP */
+/*        ,0,1,NONE,OI,OO,OO,OO,(fixnum exitc),"") */
+/* {	return fLbye(exitc); } */
  
 
 /*  c_trace(void) */
@@ -640,13 +696,13 @@ DEFUN_NEW("EXIT",object,fLexit,LISP
 /*  } */
 
 static void
-siLargc(void) {
+FFN(siLargc)(void) {
   check_arg(0);
   vs_push(make_fixnum(ARGC));
 }
 
 static void
-siLargv(void) {
+FFN(siLargv)(void) {
   int i=0;
   
   check_arg(1);
@@ -660,7 +716,7 @@ siLargv(void) {
 
 #ifdef UNIX
 static void
-siLgetenv(void) {
+FFN(siLgetenv)(void) {
 
   char name[256];
   int i;
@@ -690,14 +746,14 @@ siLgetenv(void) {
 object *vs_marker;
 
 static void
-siLmark_vs(void) {
+FFN(siLmark_vs)(void) {
   check_arg(0);
   vs_marker = vs_base;
   vs_base[0] = Cnil;
 }
 
 static void
-siLcheck_vs(void) {
+FFN(siLcheck_vs)(void) {
   check_arg(0);
   if (vs_base != vs_marker)
     FEerror("Value stack is flawed.", 0);
@@ -705,7 +761,7 @@ siLcheck_vs(void) {
 }
 
 static object
-siLcatch_fatal(int i) {
+FFN(siLcatch_fatal)(int i) {
   catch_fatal=i;
   return Cnil;
 }
@@ -720,8 +776,7 @@ reset_cstack_limit(int arg) {
 #endif
 }
 
-void
-siLreset_stack_limits(void)
+LFD(siLreset_stack_limits)(void)
 {
   int i=0;
 
@@ -787,15 +842,14 @@ multiply_stacks(int m) {
 
 DEFVAR("*NO-INIT*",sSAno_initA,SI,Cnil,"");
 
-void
-siLinit_system(void) {
+LFD(siLinit_system)(void) {
   check_arg(0);
-  init_system(sSAno_initA);
+  gcl_init_system(sSAno_initA);
   vs_base[0] = Cnil;
 }
 
 static void
-siLuser_init(void) {
+FFN(siLuser_init)(void) {
   check_arg(0);
   sLApackageA->s.s_dbind = user_package;
   user_init();
@@ -803,19 +857,19 @@ siLuser_init(void) {
 }
 
 static void
-siLaddress(void) {
+FFN(siLaddress)(void) {
   check_arg(1);
   vs_base[0] = make_fixnum((long)vs_base[0]);
 }
 
 static void
-siLnani(void) {
+FFN(siLnani)(void) {
   check_arg(1);
   vs_base[0] = (object)fixint(vs_base[0]);
 }
 
 static void
-siLinitialization_failure(void) {
+FFN(siLinitialization_failure)(void) {
   check_arg(0);
   printf("lisp initialization failed\n");
   exit(0);
@@ -828,6 +882,18 @@ DEFUNO_NEW("IDENTITY",object,fLidentity,LISP
   RETURN1 (x0);
 }
 
+DEFUNO_NEW("GCL-COMPILE-TIME",object,fSgcl_compile_time,SI
+       ,0,0,NONE,OO,OO,OO,OO,void,Lgcl_compile_time,(void),"")
+{
+  RETURN1 (make_simple_string(__DATE__ " " __TIME__));
+}
+
+DEFUNO_NEW("LDB1",object,fSldb1,SI
+       ,3,3,NONE,OI,II,OO,OO,void,Lldb1,(fixnum a,fixnum b, fixnum c),"")
+{
+  RETURN1 (make_fixnum(((((~(-1 << (a))) << (b)) & (c)) >> (b))));
+}
+
 DEFUN_NEW("LISP-IMPLEMENTATION-VERSION",object,fLlisp_implementation_version,LISP
        ,0,0,NONE,OO,OO,OO,OO,(void),"")
 {
@@ -837,7 +903,7 @@ DEFUN_NEW("LISP-IMPLEMENTATION-VERSION",object,fLlisp_implementation_version,LIS
 
 
 static void
-siLsave_system(void) {
+FFN(siLsave_system)(void) {
   int i;
   
 #ifdef HAVE_YP_UNBIND
@@ -876,10 +942,10 @@ siLsave_system(void) {
   
   
 /*  #endif */
-  cbgbccount = 0;
-  rbgbccount = 0;
+  cbgbccount = tm_table[t_contiguous].tm_adjgbccnt = 0;
+  rbgbccount = tm_table[t_relocatable].tm_adjgbccnt = 0;
   for (i = 0;  i < (int)t_end;  i++)
-    tm_table[i].tm_gbccount = 0;
+    tm_table[i].tm_gbccount = tm_table[i].tm_adjgbccnt = 0;
   Lsave();
   saving_system = FALSE;
   alloc_page(-(holepage+nrbpage));
@@ -920,10 +986,10 @@ init_main(void) {
   {object features;
   
 #define ADD_FEATURE(name) \
-	 features=  make_cons(make_ordinary(name),features)
+	 features=  make_cons(make_keyword(name),features)
   
-  features=    make_cons(make_ordinary("COMMON"),
-			 make_cons(make_ordinary("KCL"), Cnil));
+  features=    make_cons(make_keyword("COMMON"),
+			 make_cons(make_keyword("KCL"), Cnil));
   ADD_FEATURE("AKCL");
   ADD_FEATURE("GCL");
 #ifdef BROKEN_O4_OPT
@@ -938,14 +1004,25 @@ init_main(void) {
   ADD_FEATURE("UNIX");
 #endif	
 #endif
+
+#ifdef _WIN32
+  ADD_FEATURE("WINNT");
+  ADD_FEATURE("WIN32");
+#endif
+
 #ifdef IEEEFLOAT
   ADD_FEATURE("IEEE-FLOATING-POINT");
 #endif
 #ifdef SGC
   ADD_FEATURE("SGC");
 #endif	 
-#ifdef  ADDITIONAL_FEATURES
-  ADDITIONAL_FEATURES;
+/* #ifdef  ADDITIONAL_FEATURES */
+/*   ADDITIONAL_FEATURES; */
+/* #endif */
+  ADD_FEATURE(HOST_CPU);
+  ADD_FEATURE(HOST_KERNEL);
+#ifdef HOST_SYSTEM
+  ADD_FEATURE(HOST_SYSTEM);
 #endif
 #ifdef  BSD
   ADD_FEATURE("BSD");
@@ -973,11 +1050,22 @@ init_main(void) {
       }  }
 #endif	 
   
-  
+#ifdef HAVE_READLINE
+  ADD_FEATURE("READLINE");
+#endif
+#if defined(HAVE_LIBBFD) && !defined(USE_DLOPEN)
+  ADD_FEATURE("BFD");
+#endif
+  ADD_FEATURE("UNEXEC");
+
+#ifdef HAVE_GNU_LD
+  ADD_FEATURE("GNU-LD");
+#endif
   
   make_special("*FEATURES*",features);}
   
   make_si_function("SAVE-SYSTEM", siLsave_system);
   make_si_sfun("CATCH-FATAL",siLcatch_fatal,ARGTYPE1(f_fixnum));
+  make_si_function("WARN-VERSION",Lidentity);
   
 }
